/*
    TUIO Simulator - part of the reacTIVision project
    http://reactivision.sourceforge.net/

    Copyright (c) 2005-2007 Martin Kaltenbrunner <mkalten@iua.upf.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import com.illposed.osc.*;

public class Simulation extends JComponent implements Runnable {

	private Manager manager;
	private OSCPortOut oscPort;

	private int currentFrame = 0;	
	private long lastFrameTime = -1;
	private long lastCursorTime = -1;
	private float lastObjectSpeed = 0.0f;
	private float lastCursorSpeed = 0.0f;
	private int session_id = 0;
	private int cursor_id  = 0;

	private final float halfPi = (float)(Math.PI/2);
	private final float doublePi = (float)(Math.PI*2);

	Shape vision;	
	private Stroke normalStroke = new BasicStroke(2.0f);	
	private Stroke borderStroke = new BasicStroke(7.0f);
	private Color borderColor = new Color(0,239,0);
	private Stroke  gestureStroke = new BasicStroke(1.0f);
	
	int selObj = -1;
	int selCur = -1;
	int lastX = -1;
	int lastY = -1;
	int clickX = 0;
	int clickY = 0;
	//boolean mouseDown = false;
	Vector stickyCursors = new Vector();;
	
	boolean showName = false;
	int nameId, namex, namey;
	String objectName = "";

	private boolean running = false;
	
	int window_width = TuioSimulator.width;
	int window_height = TuioSimulator.height;
	int table_width = (int)(TuioSimulator.width/1.25);
	int table_height = (int)(TuioSimulator.height/1.25);
	int border_width = (int)(window_width-table_width)/2;
	int border_height = (int)(window_height-table_height)/2;
	
	public Simulation(Manager manager, String host, int port) {
		super();
		this.manager=manager;
		session_id = manager.objectList.size();
		
		try { oscPort = new OSCPortOut(java.net.InetAddress.getByName(host),port); }
		catch (Exception e) { oscPort = null; }
		reset(); reset(); reset();

		String cursorName="cursor16.gif";
		Point cursorPoint=new Point(0,0);
		double cw = Toolkit.getDefaultToolkit().getBestCursorSize(16,16).getWidth();	
		if (cw==32) { cursorName = "cursor32.gif"; cursorPoint = new Point(8,8); }
		if (cw==64) { cursorName = "cursor64.gif"; cursorPoint = new Point(24,24); }
	
		Image cursorImage = Toolkit.getDefaultToolkit().createImage("resources"+System.getProperty("file.separator")+cursorName);
		Cursor cursor = Toolkit.getDefaultToolkit().createCustomCursor(cursorImage,cursorPoint,"myCursor");
		setCursor(cursor);

		String  os = System.getProperty("os.name");
		if (os.equals("Mac OS X")) RepaintManager.currentManager(this).setDoubleBufferingEnabled(false);

		// init the table setup
		vision = new Area(new Rectangle2D.Float(border_width-12,border_height-12,table_width+12,table_height+12));
		

		// listens to the mouseDown event
		addMouseListener (
			new MouseAdapter () {
				public void mousePressed (MouseEvent evt) {
					mouse_pressed(evt);
				}
			}
		);

		// listens to the mouseDragged event
		addMouseMotionListener (
			new MouseMotionAdapter () {
				public void mouseDragged (MouseEvent evt) {
					mouse_dragged(evt);
				}
			}
		);
		
		// listens to the mouseReleased event
		addMouseListener (
			new MouseAdapter () {
				public void mouseReleased (MouseEvent evt) {
					mouse_released(evt);
				}
			}
		);

		// listens to the mouseMoved event
		addMouseMotionListener (
			new MouseMotionAdapter () {
				public void mouseMoved (MouseEvent evt) {
					mouse_moved(evt);
				}
			}
		);
	}

	private void sendOSC(OSCPacket packet) {
		try { oscPort.send(packet); }
		catch (java.io.IOException e) {}
	}

	private OSCMessage setMessage(int s_id, int f_id, int x, int y, float angle, float speed, float accel) {
		OSCMessage setMessage = new OSCMessage("/tuio/2Dobj");
		setMessage.addArgument("set");
		setMessage.addArgument(new Integer(s_id));
		setMessage.addArgument(new Integer(f_id));
		setMessage.addArgument(new Float((x-border_width)/(float)table_width));
		setMessage.addArgument(new Float((y-border_width)/(float)table_height));
		setMessage.addArgument(new Float(angle));
		setMessage.addArgument(new Float(Math.sqrt(speed/2)));
		setMessage.addArgument(new Float(Math.sqrt(speed/2)));
		setMessage.addArgument(new Float(0));
		setMessage.addArgument(new Float(accel));
		setMessage.addArgument(new Float(0));

		return setMessage;
	}

	private void cursorDelete() {

		OSCBundle cursorBundle = new OSCBundle();

		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dcur");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");

		OSCMessage aliveMessage = new OSCMessage("/tuio/2Dcur");
		aliveMessage.addArgument("alive");
		Enumeration cursorList = manager.cursorList.keys();
		while (cursorList.hasMoreElements()) {
			Integer s_id = (Integer)cursorList.nextElement();
			aliveMessage.addArgument(s_id);
		}
		
		OSCMessage frameMessage = new OSCMessage("/tuio/2Dcur");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(currentFrame));
		
		cursorBundle.addPacket(remoteMessage);
		cursorBundle.addPacket(aliveMessage);
		cursorBundle.addPacket(frameMessage);

		sendOSC(cursorBundle);
	}

	private OSCMessage setCursorMessage(Integer cursor_s_id)
	{
		
		
		float speedx = 0.0f;
		float speedy = 0.0f;
		float maccel = 0.0f;
		long currentCursorTime = System.currentTimeMillis();
		Vector gesture = (Vector)manager.cursorList.get(cursor_s_id);
		Point actualPoint = (Point)gesture.lastElement();
		Point lastPoint = (gesture.size()>1)? (Point)gesture.elementAt(gesture.size()-2) : actualPoint;
		
/*		if (lastCursorTime>0) {
			Vector gesture = (Vector)manager.cursorList.get(cursor_s_id);
			if (gesture.size()>1) {
				Point lastPoint = (Point)gesture.elementAt(gesture.size()-2);*/
				long dt = currentCursorTime - lastCursorTime;
				float dx = actualPoint.x - lastPoint.x;
				float dy = actualPoint.y - lastPoint.y;
				float dist = (float)Math.sqrt(dx*dx+dy*dy);
				float mspeed  = dist/dt;
				speedx = Math.abs(dx/dt);
				speedy = Math.abs(dy/dt);
				maccel = (mspeed-lastCursorSpeed)/dt;
		
				lastCursorSpeed = mspeed;
/*			}
		}*/
		
		OSCMessage setMessage = new OSCMessage("/tuio/2Dcur");
		setMessage.addArgument("set");
		setMessage.addArgument(cursor_s_id);
		setMessage.addArgument(new Float((actualPoint.x-border_width)/(float)table_width));
		setMessage.addArgument(new Float((actualPoint.y-border_height)/(float)table_height));
		setMessage.addArgument(new Float(speedx));
		setMessage.addArgument(new Float(speedy));
		setMessage.addArgument(new Float(maccel));
		
		return setMessage;
	}
	
	private void cursorMessage(int x, int y) {

		OSCBundle cursorBundle = new OSCBundle();

		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dcur");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");

		OSCMessage aliveMessage = new OSCMessage("/tuio/2Dcur");
		aliveMessage.addArgument("alive");

		Enumeration cursorList = manager.cursorList.keys();
		while (cursorList.hasMoreElements()) {
			Integer s_id = (Integer)cursorList.nextElement();
			aliveMessage.addArgument(s_id);
		}

		float speedx = 0.0f;
		float speedy = 0.0f;
		float maccel = 0.0f;
		long currentCursorTime = System.currentTimeMillis();
		if (lastCursorTime>0) {
			Vector gesture = (Vector)manager.cursorList.get(new Integer(selCur));
			if (gesture.size()>1) {
				Point lastPoint = (Point)gesture.elementAt(gesture.size()-2);
				long dt = currentCursorTime - lastCursorTime;
				float dx = x - lastPoint.x;
				float dy = y - lastPoint.y;
				float dist = (float)Math.sqrt(dx*dx+dy*dy);
				float mspeed  = dist/dt;
				speedx = Math.abs(dx/dt);
				speedy = Math.abs(dy/dt);
				maccel = (mspeed-lastCursorSpeed)/dt;
		
				lastCursorSpeed = mspeed;
			}
		}
		lastCursorTime = currentCursorTime;

		OSCMessage setMessage = new OSCMessage("/tuio/2Dcur");
		setMessage.addArgument("set");
		setMessage.addArgument(new Integer(selCur));
		setMessage.addArgument(new Float((x-border_width)/(float)table_width));
		setMessage.addArgument(new Float((y-border_height)/(float)table_height));
		setMessage.addArgument(new Float(speedx));
		setMessage.addArgument(new Float(speedy));
		setMessage.addArgument(new Float(maccel));

		
		
		OSCMessage frameMessage = new OSCMessage("/tuio/2Dcur");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(currentFrame));
		
		
		cursorBundle.addPacket(remoteMessage);
		cursorBundle.addPacket(aliveMessage);
		cursorBundle.addPacket(setMessage);
		cursorBundle.addPacket(frameMessage);

		if (manager.verbose) {
			System.out.println("set cur "+selCur+" "+(x-border_width)/(float)table_width+" "+(y-border_height)/(float)table_height+" "+speedx+" "+speedy+" "+maccel);
		}
		sendOSC(cursorBundle);
	}

	private void setMessageBundle(int s_id, int f_id, int x, int y, float angle, float speed, float accel) {


		OSCBundle oscBundle = new OSCBundle();

		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dobj");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");

		OSCMessage aliveMessage = new OSCMessage("/tuio/2Dobj");
		aliveMessage.addArgument("alive");

		Enumeration objectList = manager.objectList.elements();
		while (objectList.hasMoreElements()) {
			Tangible tangible = (Tangible)objectList.nextElement();
			if (tangible.isActive())
				aliveMessage.addArgument(new Integer(tangible.session_id));
		}

		OSCMessage setMessage = setMessage(s_id, f_id, x, y, angle ,speed, accel);
		if (manager.verbose) System.out.println("set obj "+s_id+" "+f_id+" "+(x-border_width)/(float)table_width+" "+(y-border_height)/(float)table_height+" "+angle+" "+Math.sqrt(speed/2)+" "+Math.sqrt(speed/2)+" "+0.0f+" "+accel+" "+0.0f);

		currentFrame++;
		OSCMessage frameMessage = new OSCMessage("/tuio/2Dobj");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(currentFrame));

		oscBundle.addPacket(remoteMessage);
		oscBundle.addPacket(aliveMessage);
		oscBundle.addPacket(setMessage);
		oscBundle.addPacket(frameMessage);
		
		sendOSC(oscBundle);
	}

	public void aliveMessage() {

		OSCBundle oscBundle = new OSCBundle();

		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dobj");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");

		OSCMessage aliveMessage = new OSCMessage("/tuio/2Dobj");
		aliveMessage.addArgument("alive");

		Enumeration objectList = manager.objectList.elements();
		while (objectList.hasMoreElements()) {
			Tangible tangible = (Tangible)objectList.nextElement();
			if (tangible.isActive())
				aliveMessage.addArgument(new Integer(tangible.session_id));
		}
		
		currentFrame++;
		OSCMessage frameMessage = new OSCMessage("/tuio/2Dobj");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(currentFrame));

		oscBundle.addPacket(remoteMessage);
		oscBundle.addPacket(aliveMessage);
		oscBundle.addPacket(frameMessage);
		
		sendOSC(oscBundle);
	}

	private void completeMessage() {

		Vector messageList = new Vector(manager.objectList.size());

		OSCMessage frameMessage = new OSCMessage("/tuio/2Dobj");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(-1));
	
		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dobj");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");	

		OSCMessage aliveMessage = new OSCMessage("/tuio/2Dobj");
		aliveMessage.addArgument("alive");

		Enumeration objectList = manager.objectList.elements();
		while (objectList.hasMoreElements()) {
			Tangible tangible = (Tangible)objectList.nextElement();
			try {
				if (tangible.isActive()) {
					int s_id = tangible.session_id;
					int f_id = tangible.fiducial_id;
					int x = (int)(tangible.getPosition().getX());
					int y = (int)(tangible.getPosition().getY());
					float a =tangible.getAngle();
					float speed = tangible.speed;
					float accel =tangible.accel;
					messageList.addElement(setMessage(s_id, f_id, x, y, a, speed, accel));
					aliveMessage.addArgument(new Integer(tangible.session_id));
				}
			} catch (Exception e) { return; }
		}

		int i;
		for (i=0;i<(messageList.size()/5);i++) {
			OSCBundle oscBundle = new OSCBundle();
	
			oscBundle.addPacket(remoteMessage);
			oscBundle.addPacket(aliveMessage);

			for (int j=0;j<5;j++)
				oscBundle.addPacket((OSCPacket)messageList.elementAt(i*5+j));
	
			if (messageList.size()<=5) oscBundle.addPacket(frameMessage);
			sendOSC(oscBundle);
		} 

		if ((messageList.size()%5!=0) || (messageList.size()==0)) {
			OSCBundle oscBundle = new OSCBundle();
		
			oscBundle.addPacket(remoteMessage);
			oscBundle.addPacket(aliveMessage);
	
			for (int j=0;j<messageList.size()%5;j++)
				oscBundle.addPacket((OSCPacket)messageList.elementAt(i*5+j));	
	
			oscBundle.addPacket(frameMessage);
			sendOSC(oscBundle);
		}
	}
	
	// ADDED TO RESEND ALIVE MESSAGES FOR CURSORS!
	private void complete2DCurMessage() {
	
		Vector messageList = new Vector(manager.cursorList.size());
		OSCMessage frameMessage = new OSCMessage("/tuio/2Dcur");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(-1));
		
		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dcur");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");
		
		OSCMessage aliveMessage = new OSCMessage("/tuio/2Dcur");
		aliveMessage.addArgument("alive");
		
		Enumeration cursorsKeys =manager.cursorList.keys();
		while(cursorsKeys.hasMoreElements())
		{
			Integer s_id = (Integer)cursorsKeys.nextElement();
			messageList.addElement(setCursorMessage(s_id));
			aliveMessage.addArgument(s_id);
		}
		
		int i;
		for (i=0;i<(messageList.size()/5);i++) {
			OSCBundle oscBundle = new OSCBundle();
	
			oscBundle.addPacket(remoteMessage);
			oscBundle.addPacket(aliveMessage);

			for (int j=0;j<5;j++)
				oscBundle.addPacket((OSCPacket)messageList.elementAt(i*5+j));
	
			if (messageList.size()<=5) oscBundle.addPacket(frameMessage);
			sendOSC(oscBundle);
		} 

		if ((messageList.size()%5!=0) || (messageList.size()==0)) {
			OSCBundle oscBundle = new OSCBundle();
		
			oscBundle.addPacket(remoteMessage);
			oscBundle.addPacket(aliveMessage);
	
			for (int j=0;j<messageList.size()%5;j++)
				oscBundle.addPacket((OSCPacket)messageList.elementAt(i*5+j));	
	
			oscBundle.addPacket(frameMessage);
			sendOSC(oscBundle);
		}
		
	}

	public void quit() {
		running = false;
		reset(); reset(); reset();
	}

	public void reset() {
		session_id = manager.objectList.size();
		cursor_id=0;
		stickyCursors.clear();
		
		lastFrameTime = -1;
		lastCursorTime = -1;
		
		OSCBundle oscBundle = new OSCBundle();

		OSCMessage remoteMessage = new OSCMessage("/tuio/2Dobj");
		remoteMessage.addArgument("source");
		remoteMessage.addArgument("vision");

		OSCMessage objectAliveMessage = new OSCMessage("/tuio/2Dobj");
		objectAliveMessage.addArgument("alive");

		OSCMessage cursorAliveMessage = new OSCMessage("/tuio/2Dcur");
		cursorAliveMessage.addArgument("alive");


		currentFrame++;
		OSCMessage frameMessage = new OSCMessage("/tuio/2Dobj");
		frameMessage.addArgument("fseq");
		frameMessage.addArgument(new Integer(currentFrame));

		oscBundle.addPacket(remoteMessage);
		oscBundle.addPacket(objectAliveMessage);
		oscBundle.addPacket(cursorAliveMessage);
		oscBundle.addPacket(frameMessage);
		
		sendOSC(oscBundle);
	}

	public void paint(Graphics g) {
		update(g);
	}

	public void update(Graphics g) {
	
		// setup the graphics environment
		Graphics2D g2 = (Graphics2D)g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
		g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);
		g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);

		// draw the table & border
		g2.setPaint(Color.gray);
		g2.fillRect(0,0,window_width,window_height);
		g2.setPaint(Color.white);
		g2.fillRect(border_width,border_height,table_width,table_height);

		//draw the round table edge
		g2.setPaint(Color.lightGray);
		g2.drawOval((int)((window_width-table_height)/2),border_height,table_height,table_height);

		// paint the cursors
		Enumeration cursorList = manager.cursorList.elements();
		while (cursorList.hasMoreElements()) {
			Vector gesture = (Vector)cursorList.nextElement();
			if (gesture.size()>0) {
				g2.setPaint(Color.blue);
				g2.setStroke(gestureStroke);
				Point start = (Point)(gesture.elementAt(0));
				for (int i=0;i<gesture.size();i++) {
					Point end  = (Point)(gesture.elementAt(i));
					g2.draw(new Line2D.Double(start.getX(),start.getY(),end.getX(),end.getY()));
					start = end;
				}
				g2.setPaint(Color.lightGray);
				g2.fillOval((int)start.getX()-5,(int)start.getY()-5,10,10);
			}
		}

		// paint the objects
		Enumeration objectList = manager.objectList.elements();
		while (objectList.hasMoreElements()) {
			Tangible tangible = (Tangible)objectList.nextElement();

			// draw the border
			if (tangible.isActive()) {
				g2.setStroke(borderStroke);
				g2.setPaint(borderColor);
				g2.draw(tangible.geom);
			}

			// draw the objects
			g2.setPaint(tangible.color);
			g2.fill(tangible.geom);

			// draw the rotation mark
			g2.setPaint(Color.black);
			g2.setStroke(normalStroke);
			g2.draw(new Line2D.Float(tangible.getPosition(),tangible.getPointer()));
			if ((tangible.previous!=null)  && (tangible.next!=null))
				g2.fillOval((int)tangible.getPosition().getX()-3,(int)tangible.getPosition().getY()-3,6,6);

			// draw the object id
			g2.setPaint(Color.white);
			g2.drawString(tangible.fiducial_id+"",tangible.getPosition().x+2,tangible.getPosition().y);
		}

		// draw the object name
		if (showName) {
			int boxwidth = getFontMetrics(getFont()).stringWidth(objectName)+12;
			g2.setPaint(Color.white);
			g.fillRect(namex,namey,boxwidth,16);
			g2.setPaint(Color.blue);
			g.drawRect(namex,namey,boxwidth,16);
			g2.drawString(objectName,namex+6,namey+12);
		}
	}
 
	public final void switchSide(int old_id, int increment) {
	
		Tangible tangible = (Tangible)manager.objectList.get(new Integer(old_id));
		if ((tangible.previous!=null) && (tangible.previous!=null)) {
							
			int x_save = (int)(tangible.getPosition().getX());
			int y_save = (int)(tangible.getPosition().getY());
			float a_save =  tangible.getAngle();

			Tangible new_side = tangible;
			if (increment>0) new_side = tangible.next;
			else  if (increment<0) new_side = tangible.previous;	
			selObj = new_side.session_id;

			manager.deactivate(old_id);
			manager.update(old_id,-100,-100,a_save,0.0f,0.0f);
			manager.update(selObj,x_save,y_save,a_save, 0.0f, 0.0f);
			aliveMessage();

		}
	}

	public void mouse_dragged (MouseEvent evt) {

		long currentFrameTime = System.currentTimeMillis();
		long dt = currentFrameTime -lastFrameTime;
		if (dt<20) return;

		Point pt = evt.getPoint();
		int x = (int)pt.getX();
		int y = (int)pt.getY();

		int dx = lastX -x;
		int dy = lastY -y;
		float distance = (float)(Math.sqrt(dx*dx+dy*dy));
			
		float speed = distance/dt;
		float accel = (lastObjectSpeed -speed)/dt;

		/*if (!mouseDown) {
			mouse_pressed(evt);
			mouseDown = true;
		}*/

		

		if (selObj>-1){
			Tangible selObject = (Tangible)manager.objectList.get(new Integer(selObj));
			if (selObject==null) { System.out.println(selObj); return; }
			//System.out.println(evt.getModifiers());
			switch (evt.getModifiers()) {
				// translation
				case 17: {
					// deactivate
					manager.deactivate(selObj);
					//move the object
					manager.update(selObj,x+clickX,y+clickY,selObject.getAngle(),speed,accel);
					//setMessageBundle(selObj,x,y,selObject.getAngle(),speed,accel);
					aliveMessage();

					break;
				}
				case 16: {
					//move the object
					manager.update(selObj,x+clickX,y+clickY,selObject.getAngle(),speed,accel);

					// activate
					if (vision.contains(selObject.geom.getBounds2D())) {
						if (!selObject.isActive()) {
							session_id++;
							manager.activate(selObj,session_id);
							selObj = session_id;
						}
						setMessageBundle(selObject.session_id, selObject.fiducial_id,x,y,selObject.getAngle(),speed,accel);
						//aliveMessage();
					} else { 
						manager.deactivate(selObj);
						aliveMessage();
					}
					break;
				}
				// rotation
				case 4: {
					int diff = lastY-y;
					if (diff>15) diff = 15;
					if (diff<-15) diff = -15;
					float newAngle = (float)(selObject.getAngle()+(Math.PI/90*diff));
					if(newAngle<0) newAngle+=doublePi;
					if(newAngle>doublePi) newAngle-=doublePi;
					manager.update(selObject.session_id,selObject.getPosition().x,selObject.getPosition().y,newAngle,0,0);
					if (selObject.isActive())
						setMessageBundle(selObject.session_id,selObject.fiducial_id,selObject.getPosition().x,selObject.getPosition().y,newAngle,0,0);
					break;
				}
			
				// switch sides
				case 21: //mac
				case 5: {
					int diff = lastY-y;
					if (diff>=3) switchSide(selObj,-1);
					if (diff<=-3) switchSide(selObj,1);
					break;
				}
			}
			
			} else if (selCur>-1){
				if (vision.contains(pt)) {
					manager.updateCursor(selCur,pt.x,pt.y);
					cursorMessage(pt.x,pt.y);
				} else {
					if (manager.verbose) System.out.println("del cur "+selCur);
					stickyCursors.remove(new Integer(selCur));
					manager.terminateCursor(selCur);
					cursorDelete();
					selCur = -1;
					lastCursorSpeed = 0.0f;
					lastCursorTime = -1;
					cursor_id++;				
				}
			}

			lastX = x;
			lastY = y;
			
			lastFrameTime = currentFrameTime;
			lastObjectSpeed = speed;
	}

	public void mouse_moved(MouseEvent evt) {

		if (evt.getModifiers()==2) {

			int x = evt.getX();
			int y = evt.getY();
				
			Enumeration objectList = manager.objectList.elements();
			while (objectList.hasMoreElements()) {
				Tangible tangible = (Tangible)objectList.nextElement();
				if (tangible.containsPoint(x,y)) {
	
					objectName = tangible.type.description;
					showName=true;
					nameId = tangible.fiducial_id;
					namex = x+24;
					namey = y;

					repaint();
					return;
				} 
			} 

			if (showName) {
				showName = false;
				repaint();
			}
		} else if (showName)  {
			showName = false;
			repaint();
		}
	}

	public void mouse_pressed (MouseEvent evt) {
		int x = evt.getX();
		int y = evt.getY();
				
		Enumeration objectList = manager.objectList.elements();
		while (objectList.hasMoreElements()) {
			Tangible tangible = (Tangible)objectList.nextElement();
			if (tangible.containsPoint(x,y)) {
				selObj = tangible.session_id;
				selCur = -1;
			
				clickX = (int)tangible.getPosition().getX()-x;
				clickY = (int)tangible.getPosition().getY()-y;

				lastX = x;
				lastY = y;
				return;
			}
		}

		Enumeration cursorList = manager.cursorList.keys();
		while (cursorList.hasMoreElements()) {
			Integer s_id = (Integer)cursorList.nextElement();
			Point c_point = (Point)(((Vector)manager.cursorList.get(s_id)).lastElement());
			if (c_point.distance(x,y)<7) {

				if (((evt.getModifiers() & InputEvent.SHIFT_MASK) > 0) && selCur != s_id.intValue()) {
					if (manager.verbose) System.out.println("del cur "+s_id.intValue());
					stickyCursors.remove(s_id);
					manager.terminateCursor(s_id.intValue());
					cursorDelete();
					selCur = -1;
					lastCursorSpeed = 0.0f;
					lastCursorTime = -1;
					//cursor_id++;
					return;
				} else {
					selCur = s_id.intValue();
					selObj = -1;
				
					lastX = x;
					lastY = y;
					return;
				} 
			}
		}

		if (vision.contains(new Point(x, y))) {
			selCur = cursor_id;
			if (manager.verbose) System.out.println("add cur "+selCur);
			manager.updateCursor(selCur,x,y);
			cursorMessage(x,y);
			if ((evt.getModifiers() & InputEvent.SHIFT_MASK) > 0) stickyCursors.addElement(new Integer(selCur));
			return;
		}


		selObj = -1;
		selCur = -1;
		lastX = -1;
		lastY = -1;
	}

	public void mouse_released (MouseEvent evt) {

		if ( (selObj>-1) /*&& (mouseDown)*/ )  {

			Tangible tangible = (Tangible)manager.objectList.get(new Integer(selObj));
			if (tangible==null) { System.out.println(selObj); return; }
			// activate if inside the table
			if (vision.contains(tangible.geom.getBounds2D())) {
				if (!tangible.isActive()) {
					session_id++;
					manager.activate(selObj,session_id);
					selObj = session_id;
				}
				setMessageBundle(tangible.session_id,tangible.fiducial_id,tangible.getPosition().x,tangible.getPosition().y,tangible.getAngle(),0,0);
			} else {
				manager.deactivate(selObj);
				aliveMessage();
			}

			selObj = -1;
			lastObjectSpeed = 0.0f;
		} else if ( (selCur>-1) )  {
			if (!stickyCursors.contains(new Integer(selCur))) {
				if (manager.verbose) System.out.println("del cur "+ selCur);
				manager.terminateCursor(selCur);
				cursorDelete();
			}
			selCur = -1;
			lastCursorSpeed = 0.0f;
			lastCursorTime = -1;
			cursor_id++;
		}

		//mouseDown = false;
	}

	// send table state every second
	public void run() {
		running = true;
		while(running) {
			try { Thread.sleep(1000); }
			catch (Exception e) {}
			completeMessage();
			complete2DCurMessage();
		}
	}
}
